# Generated by Django 2.0.3 on 2018-05-02 16:07

from django.db import migrations
from quartet_epcis.parsing.steps import create_rule


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.RunPython(create_rule)
    ]
